package flm.b4a.accelerview;

import android.graphics.Camera;
import android.graphics.Canvas;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.PorterDuff;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Region.Op;
import android.graphics.drawable.Drawable;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.keywords.constants.TypefaceWrapper;
import anywheresoftware.b4a.objects.drawable.CanvasWrapper.BitmapWrapper;
import anywheresoftware.b4a.objects.drawable.CanvasWrapper.RectWrapper;

@BA.ShortName("AS_Canvas")
public class AcceleratedCanvas
{
	protected Canvas cvs;
	private Paint paint;
	private RectF rectF;
	private Camera cam;
	private Matrix matrix;
	private int width;
	private int height;

	public AcceleratedCanvas()
	{
		cvs = null;
		paint = null;
		rectF = null;
		cam = null;
		matrix = null;
		width = 0;
		height = 0;
	}

	protected void Initialize(Canvas c, int w, int h, Camera cmr, Paint p, Matrix m, RectF r)
	{
		cvs = c;
		paint = p;
		rectF = r;
		cam = cmr;
		matrix = m;
		width = w;
		height = h;
	}

	/*-----------------------------------------------------------------
	  MATRIX
	 -----------------------------------------------------------------*/

	/**
	 * Imports a Matrix from another library.
	 */
	public void ImportMatrix(Matrix m)
	{
		matrix = m;
	}

	//	/**
	//	 * Resets the internal matrix.
	//	 */
	//	public void MatrixReset()
	//	{
	//		matrix.reset();
	//	}

	//	/**
	//	 * Preconcats the internal matrix with the specified rotation.
	//	 */
	//	public AcceleratedCanvas MatrixPreRotate(float degrees)
	//	{
	//		matrix.preRotate(degrees);
	//		return this;
	//	}
	//
	//	/**
	//	 * Preconcats the internal matrix with the specified rotation.
	//	 * @param px,py The coordinates of the pivot point.
	//	 */
	//	public AcceleratedCanvas MatrixPreRotate2(float degrees, float px, float py)
	//	{
	//		matrix.preRotate(degrees, px, py);
	//		return this;
	//	}

	/**
	 * Postconcats the internal matrix with the specified rotation.
	 */
	public void MatrixPostRotate(float degrees)
	{
		matrix.postRotate(degrees);
	}

	/**
	 * Postconcats the internal matrix with the specified rotation.
	 * @param px,py The coordinates of the pivot point.
	 */
	public void MatrixPostRotate2(float degrees, float px, float py)
	{
		matrix.postRotate(degrees, px, py);
	}

	/**
	 * Sets the internal matrix to rotate around (0,0) by the specified number of degrees.
	 */
	public AcceleratedCanvas MatrixSetRotate(float degrees)
	{
		matrix.setRotate(degrees);
		return this;
	}

	/**
	 * Sets the internal matrix to rotate by the specified number of degrees, with a pivot point at (px, py).
	 */
	public AcceleratedCanvas MatrixSetRotate2(float degrees, float px, float py)
	{
		matrix.setRotate(degrees, px, py);
		return this;
	}

	//	/**
	//	 * Preconcats the internal matrix with the specified scale.
	//	 */
	//	public AcceleratedCanvas MatrixPreScale(float sx, float sy)
	//	{
	//		matrix.preScale(sx, sy);
	//		return this;
	//	}

	/**
	 * Postconcats the internal matrix with the specified scale.
	 */
	public void MatrixPostScale(float sx, float sy)
	{
		matrix.postScale(sx, sy);
	}

	/**
	 * Sets the internal matrix to scale by sx and sy.
	 */
	public AcceleratedCanvas MatrixSetScale(float sx, float sy)
	{
		matrix.setScale(sx, sy);
		return this;
	}

	/**
	 * Sets the internal matrix to rotate by the specified sine and cosine values, with a pivot point at (px, py).
	 * The pivot point is the coordinate that should remain unchanged by the specified transformation.
	 */
	public AcceleratedCanvas MatrixSetSinCos(float sinValue, float cosValue, float px, float py)
	{
		matrix.setSinCos(sinValue, cosValue, px, py);
		return this;
	}

	//	/**
	//	 * Preconcats the internal matrix with the specified skew.
	//	 */
	//	public AcceleratedCanvas MatrixPreSkew(float sx, float sy)
	//	{
	//		matrix.preSkew(sx, sy);
	//		return this;
	//	}

	/**
	 * Postconcats the internal matrix with the specified skew.
	 */
	public void MatrixPostSkew(float sx, float sy)
	{
		matrix.postSkew(sx, sy);
	}

	/**
	 * Sets the internal matrix to skew by sx and sy.
	 */
	public AcceleratedCanvas MatrixSetSkew(float sx, float sy)
	{
		matrix.setSkew(sx, sy);
		return this;
	}

	/**
	 * Sets the internal matrix to skew by sx and sy, with a pivot point at (px, py).
	 */
	public AcceleratedCanvas MatrixSetSkew2(float sx, float sy, float px, float py)
	{
		matrix.setSkew(sx, sy, px, py);
		return this;
	}

	//	/**
	//	 * Preconcats the internal matrix with the specified translation.
	//	 */
	//	public AcceleratedCanvas MatrixPreTranslate(float dx, float dy)
	//	{
	//		matrix.preTranslate(dx, dy);
	//		return this;
	//	}

	/**
	 * Postconcats the internal matrix with the specified translation.
	 */
	public void MatrixPostTranslate(float dx, float dy)
	{
		matrix.postTranslate(dx, dy);
	}

	/**
	 * Sets the internal matrix to translate by (dx, dy).
	 */
	public AcceleratedCanvas MatrixSetTranslate(float dx, float dy)
	{
		matrix.setTranslate(dx, dy);
		return this;
	}

	/**
	 * Transforms the given bitmap with the internal matrix and draws it at the specified position.
	 * @param Filter Filtering affects the sampling of the bitmap when it is transformed. True = best quality.
	 */
	public void DrawBitmapWithMatrixAt(BitmapWrapper Bmp, int Left, int Top, boolean Filter)
	{
		matrix.postTranslate(Left, Top);
		if (Filter) {
			paint.reset();
			setPaintFilter();
			cvs.drawBitmap(Bmp.getObject(), matrix, paint);
		}
		else
			cvs.drawBitmap(Bmp.getObject(), matrix, null);
	}

	/**
	 * Transforms the given bitmap object with the internal matrix and draws it at the specified position.
	 */
	public void DrawBitmapObjectWithMatrixAt(ComplexObject objBmp, int Left, int Top) throws Exception
	{
		if (objBmp.objType == ComplexObject.TYPE_BITMAP || objBmp.objType == ComplexObject.TYPE_BITMAP2) {
			matrix.postTranslate(Left, Top);
			cvs.drawBitmap(objBmp.bmpParam, matrix, objBmp.paint);
		}
		else
			throw(new Exception("DrawBitmapObjectWithMatrixAt: wrong object type"));
	}

	/*-----------------------------------------------------------------
	  DRAW METHODS
	 -----------------------------------------------------------------*/

	private void ResetWithColor(int color)
	{
		paint.reset();
		paint.setColor(color);
	}

	private void setPaintFilter()
	{
		paint.setAntiAlias(true);
		paint.setFilterBitmap(true);
	}

	/**
	 * Draws the given bitmap.
	 * @param SrcRect The subset of the bitmap that will be drawn. If Null then the complete bitmap will be drawn.
	 * @param DestRect The rectangle that the bitmap will be drawn into.
	 */
	public void DrawBitmap(BitmapWrapper Bmp, Rect SrcRect, Rect DestRect)
	{
		cvs.drawBitmap(Bmp.getObject(), SrcRect, DestRect, null);
	}

	/**
	 * Draws the given bitmap at the specified position.
	 */
	public void DrawBitmapAt(BitmapWrapper Bmp, int Left, int Top)
	{
		cvs.drawBitmap(Bmp.getObject(), Left, Top, null);
	}

	/**
	 * Draws the given bitmap through a mesh.
	 * The mesh vertices are evenly distributed across the bitmap. There are MeshWidth+1 vertices across, and MeshHeight+1 vertices down. The Verts array is accessed in row-major order, so that the first MeshWidth+1 vertices are distributed across the top of the bitmap from left to right.
	 * @param MeshWidth The number of columns in the mesh. Nothing is drawn if this is 0.
	 * @param MeshHeight The number of rows in the mesh. Nothing is drawn if this is 0.
	 * @param Verts Array of x,y pairs, specifying where the mesh should be drawn. There must be at least (MeshWidth+1)*(MeshHeight+1)*2+VertOffset values in the array.
	 * @param VertOffset Number of Verts elements to skip before drawing.
	 */
	public void DrawBitmapMesh(BitmapWrapper Bmp, int MeshWidth, int MeshHeight, float[] Verts, int VertOffset) throws Exception
	{
		if (Bmp == null || !Bmp.IsInitialized()) {
			throw(new Exception("DrawBitmapMesh: bitmap not initialized"));
		}
		// The hardware acceleration does not support colors so they are skipped
		cvs.drawBitmapMesh(Bmp.getObject(), MeshWidth, MeshHeight, Verts, VertOffset, null, 0, null);
	}

	/**
	 * Draws the specified arc, which will be scaled to fit inside the specified oval.
	 * @param OvalRect The bounds of oval used to define the shape and size of the arc.
	 * @param StartAngle Starting angle (in degrees) where the arc begins.
	 * @param SweepAngle Sweep angle (in degrees) measured clockwise.
	 * @param Wedge Whether the center of the oval closes the arc to draw a wedge.
	 * @param Filled Whether the arc will be filled.
	 * @param StrokeWidth The stroke width. Relevant when Filled = False.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public void DrawArc(RectWrapper OvalRect, float StartAngle, float SweepAngle, boolean Wedge, int Color, boolean Filled, float StrokeWidth, boolean Filter)
	{
		ResetWithColor(Color);
		if (Filter) setPaintFilter();
		paint.setStyle(Filled ? Paint.Style.FILL : Paint.Style.STROKE);
		paint.setStrokeWidth(StrokeWidth);
		rectF.set(OvalRect.getObject());
		cvs.drawArc(rectF, StartAngle, SweepAngle, Wedge, paint);
	}

	/**
	 * Draws a circle.
	 * @param x The x-coordinate of the center of the circle to be drawn.
	 * @param y The y-coordinate of the center of the circle to be drawn.
	 * @param Radius The radius of the circle to be drawn.
	 * @param Filled Whether the circle will be filled.
	 * @param StrokeWidth The stroke width. Relevant when Filled = False.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public void DrawCircle(float x, float y, float Radius, int Color, boolean Filled, float StrokeWidth, boolean Filter)
	{
		ResetWithColor(Color);
		if (Filter) setPaintFilter();
		paint.setStyle(Filled ? Paint.Style.FILL : Paint.Style.STROKE);
		paint.setStrokeWidth(StrokeWidth);
		cvs.drawCircle(x, y, Radius, paint);
	}

	/**
	 * Fills the entire canvas with the given color.
	 */
	public void DrawColor(int Color)
	{
		if (Color == 0) // Transparent color -> pixels are erased
			cvs.drawColor(Color, PorterDuff.Mode.CLEAR);
		else
			cvs.drawColor(Color);
	}

	/**
	 * Draws a Drawable into the specified rectangle.
	 */
	public void DrawDrawable(Drawable drawable, RectWrapper DestRect)
	{
		drawable.setBounds(DestRect.getObject());
		drawable.draw(cvs);
	}

	/**
	 * Draws a line from (x1, y1) to (x2, y2).
	 * @param StrokeWidth Width of the line.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public void DrawLine(float x1, float y1, float x2, float y2, int Color, float StrokeWidth, boolean Filter)
	{
		ResetWithColor(Color);
		if (Filter) setPaintFilter();
		paint.setStrokeWidth(StrokeWidth);
		cvs.drawLine(x1, y1, x2, y2, paint);
	}

	/**
	 * Draws an AS_Object.
	 * @param X,Y For circles, X and Y define the position of the centre. For other objects, they define the position of the upper left corner.
	 */
	public void DrawObjectAt(ComplexObject Obj, float X, float Y) throws Exception
	{
		if (Obj == null || !Obj.getIsInitialized()) 
			throw(new Exception("DrawObjectAt: object not initialized"));
		switch (Obj.objType) {
		case ComplexObject.TYPE_ARC:
			rectF.set(X, Y, X + Obj.iWidth, Y + Obj.iHeight);
			cvs.drawArc(rectF, Obj.fParam1, Obj.fParam2, Obj.boolParam, Obj.paint);
			break;
		case ComplexObject.TYPE_BITMAP:
			cvs.drawBitmap(Obj.bmpParam, X, Y, Obj.paint);
			break;
		case ComplexObject.TYPE_BITMAP2:
			rectF.set(X, Y, X + Obj.iWidth, Y + Obj.iHeight);
			cvs.drawBitmap(Obj.bmpParam, Obj.rctParam, rectF, Obj.paint);
			break;
		case ComplexObject.TYPE_CIRCLE:
			cvs.drawCircle(X, Y, Obj.fParam1, Obj.paint);
			break;
		case ComplexObject.TYPE_LINE:
			cvs.drawLine(X, Y, X + Obj.fParam1, Y + Obj.fParam2, Obj.paint);
			break;
		case ComplexObject.TYPE_OVAL:
			rectF.set(X, Y, X + Obj.iWidth, Y + Obj.iHeight);
			cvs.drawOval(rectF, Obj.paint);
			break;
		case ComplexObject.TYPE_PATH:
			if (X != 0 || Y != 0) {
				cvs.save();
				cvs.translate(X, Y);
				cvs.drawPath(Obj.pathParam, Obj.paint);
				cvs.restore();
			}
			else
				cvs.drawPath(Obj.pathParam, Obj.paint);
			break;
		case ComplexObject.TYPE_POINT:
			cvs.drawPoint(X, Y, Obj.paint);
			break;
		case ComplexObject.TYPE_RECT:
			rectF.set(X, Y, X + Obj.iWidth, Y + Obj.iHeight);
			cvs.drawRect(rectF, Obj.paint);
			break;
		case ComplexObject.TYPE_ROUNDRECT:
			rectF.set(X, Y, X + Obj.iWidth, Y + Obj.iHeight);
			cvs.drawRoundRect(rectF, Obj.fParam1, Obj.fParam2, Obj.paint);
			break;
		}
	}

	/**
	 * Draws an oval shape.
	 * @param Rect The rectangle bounds of the oval to be drawn.
	 * @param Filled Whether the oval will be filled.
	 * @param StrokeWidth The stroke width. Relevant when Filled = False.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public void DrawOval(RectWrapper Rect, int Color, boolean Filled, float StrokeWidth, boolean Filter)
	{
		ResetWithColor(Color);
		if (Filter) setPaintFilter();
		paint.setStyle(Filled ? Paint.Style.FILL : Paint.Style.STROKE);
		paint.setStrokeWidth(StrokeWidth);
		rectF.set(Rect.getObject());
		cvs.drawOval(rectF, paint);
	}

	/**
	 * Draws the given path.
	 * @param Path The path to be drawn. It may be an instance of the Path class or an instance of the AS_Path class.
	 * @param Filled Whether the shape bounded by the path will be filled.
	 * @param StrokeWidth The stroke width. Relevant when Filled = False.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public void DrawPath(Object Path, int Color, boolean Filled, float StrokeWidth, boolean Filter) throws Exception
	{
		ResetWithColor(Color);
		if (Filter) setPaintFilter();
		paint.setStyle(Filled ? Paint.Style.FILL : Paint.Style.STROKE);
		paint.setStrokeWidth(StrokeWidth);
		if (Path.getClass().getName().equals("android.graphics.Path"))
			cvs.drawPath((android.graphics.Path) Path, paint);
		else if (Path instanceof ComplexPath)
			cvs.drawPath(((ComplexPath)Path).path, paint);
		else
			throw(new Exception("DrawPath: invalid path"));
	}

	/**
	 * Draws a point at the specified position and color.
	 */
	public void DrawPoint(float x, float y, int Color)
	{
		ResetWithColor(Color);
		paint.setStyle(Paint.Style.STROKE);
		paint.setStrokeWidth(0.0F);
		cvs.drawPoint(x, y, paint);
	}

	/**
	 * Draws a rectangle.
	 * @param Rect The coordinates of the rectangle to be drawn.
	 * @param Filled Whether the rectangle will be filled.
	 * @param StrokeWidth The stroke width. Relevant when Filled = False.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public void DrawRect(RectWrapper Rect, int Color, boolean Filled, float StrokeWidth, boolean Filter)
	{
		ResetWithColor(Color);
		if (Filter) setPaintFilter();
		paint.setStyle(Filled ? Paint.Style.FILL : Paint.Style.STROKE);
		paint.setStrokeWidth(StrokeWidth);
		cvs.drawRect(Rect.getObject(), paint);
	}

	/**
	 * Draws a rectangle with rounded corners.
	 * @param Rect The coordinates of the round-rect to be drawn.
	 * @param CornerRadius The radius of the oval used to round the corners.
	 * @param Filled Whether the round-rect will be filled.
	 * @param StrokeWidth The stroke width. Relevant when Filled = False.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public void DrawRoundRect(RectWrapper Rect, int Color, float CornerRadius, boolean Filled, float StrokeWidth, boolean Filter)
	{
		ResetWithColor(Color);
		if (Filter) setPaintFilter();
		paint.setStyle(Filled ? Paint.Style.FILL : Paint.Style.STROKE);
		paint.setStrokeWidth(StrokeWidth);
		rectF.set(Rect.getObject());
		cvs.drawRoundRect(rectF, CornerRadius, CornerRadius, paint);
	}

	public final static Paint.Align ALIGN_LEFT = Paint.Align.LEFT;
	public final static Paint.Align ALIGN_CENTER = Paint.Align.CENTER;
	public final static Paint.Align ALIGN_RIGHT = Paint.Align.RIGHT;

	/**
	 * Draws the text.
	 * @param Text The text to be drawn.
	 * @param x,y The origin point.
	 * @param Typeface Typeface (font) to use.
	 * @param TextSize Size of text. This value is automatically scaled so do not scale it yourself.
	 * @param Align The alignment related to the origin. One of the ALIGN constants.
	 */
	public void DrawText(String Text, float x, float y, TypefaceWrapper Typeface, float TextSize, int Color, Paint.Align Align)
	{
		ResetWithColor(Color);
		setPaintFilter();
		paint.setStrokeWidth(0.0F);
		paint.setStyle(Paint.Style.FILL);
		paint.setTextAlign(Align);
		paint.setTextSize(TextSize * BA.applicationContext.getResources().getDisplayMetrics().scaledDensity);
		paint.setTypeface(Typeface.getObject());
		cvs.drawText(Text, x, y, paint);
	}

	/**
	 * Draws an AS_Text object.
	 * @param Text The text to be drawn.
	 * @param X,Y The origin point.
	 * @param Align The alignment related to the origin. One of the ALIGN constants.
	 */
	public void DrawTextObject(ComplexText objText, String Text, float X, float Y, Paint.Align Align) throws Exception
	{
		if (objText == null || !objText.getIsInitialized())
			throw(new Exception("DrawTextObject: object not initialized"));
		objText.getCO().paint.setTextAlign(Align);
		if (objText.getCO().pathParam == null)
			cvs.drawText(Text, X, Y, objText.getCO().paint);
		else
			cvs.drawTextOnPath(Text, objText.getCO().pathParam, X, Y, objText.getCO().paint);
	}

	/**
	 * Draws the text along the specified path.
	 * Warning: This does not work under Honeycomb or ICS if the hardware acceleration is enabled.
	 * @param Text The text to be drawn.
	 * @param x1,y1 The coordinates of the start point.
	 * @param x2,y2,x3,y3 The coordinates of the intermediate points.
	 * @param x4,y4 The coordinates of the end point.
	 * @param Typeface Typeface (font) to use.
	 * @param TextSize Size of text. This value is automatically scaled so do not scale it yourself.
	 * @param Align The alignment related to the origin. One of the ALIGN constants.
	 */
	public void DrawTextOnPath(String Text, float x1, float y1, float x2, float y2, float x3, float y3, float x4, float y4, TypefaceWrapper Typeface, float TextSize, int Color, Paint.Align Align)
	{
		ResetWithColor(Color);
		setPaintFilter();
		paint.setStrokeWidth(0.0F);
		paint.setStyle(Paint.Style.FILL);
		paint.setTextAlign(Align);
		paint.setTextSize(TextSize * BA.applicationContext.getResources().getDisplayMetrics().scaledDensity);
		paint.setTypeface(Typeface.getObject());
		Path path = new Path();
		path.moveTo(x1, y1);
		path.cubicTo(x2, y2, x3, y3, x4, y4);
		cvs.drawTextOnPath(Text, path, 0, 0, paint);
	}

	/**
	 * Returns the width of the given text.
	 */
	public float MeasureStringWidth(String Text, TypefaceWrapper Typeface, float TextSize)
	{
		paint.reset();
		setPaintFilter();
		paint.setStrokeWidth(0.0F);
		paint.setStyle(Paint.Style.FILL);
		paint.setTextAlign(ALIGN_LEFT);
		paint.setTextSize(TextSize * BA.applicationContext.getResources().getDisplayMetrics().scaledDensity);
		paint.setTypeface(Typeface.getObject());
		return paint.measureText(Text);
	}

	/**
	 * Returns the height of the given text.
	 */
	public int MeasureStringHeight(String Text, TypefaceWrapper Typeface, float TextSize)
	{
		paint.reset();
		setPaintFilter();
		paint.setStrokeWidth(0.0F);
		paint.setStyle(Paint.Style.FILL);
		paint.setTextAlign(ALIGN_LEFT);
		paint.setTextSize(TextSize * BA.applicationContext.getResources().getDisplayMetrics().scaledDensity);
		paint.setTypeface(Typeface.getObject());
		Rect r = new Rect();
		paint.getTextBounds(Text, 0, Text.length(), r);
		return r.height();
	}

	/*-----------------------------------------------------------------
	  SAVE/RESTORE
	 -----------------------------------------------------------------*/

	/**
	 * Saves the current state of the canvas. Each SaveState should be balanced with a call to RestoreState.
	 *
	 * In this example of rotation, only the rectangle is rotated:
	 * AC.SaveState
	 * AC.RotateCanvas(45)
	 * AC.DrawRect(rctRect, Colors.Red, True, 0, True)
	 * AC.RestoreState
	 * AC.DrawOval(rctOval, Colors.Blue, True, 0, True)
	 */
	public void SaveState()
	{
		cvs.save();
	}

	/**
	 * Restores the canvas state saved with SaveState.
	 * Undoes the following functions called after SaveState:
	 * RotateCanvas, RotateCanvasAround, ScaleCanvas, SkewCanvas, TranslateCanvas, ClipPath, and ClipRect.
	 * The drawings stay unchanged.
	 */
	public void RestoreState()
	{
		cvs.restore();
	}

	private int SL = -1;

	/**
	 * Redirects all subsequent drawings in a separate layer until you call TransferLayer.
	 * @param LayerRect Rectangular coordinates of the layer.
	 */
	public int CreateLayer(RectWrapper LayerRect)
	{
		rectF.set(LayerRect.getObject());
		if (SL == -1)
			SL = cvs.saveLayer(rectF, null, Canvas.ALL_SAVE_FLAG);
		else
			SL += cvs.saveLayer(rectF, null, Canvas.ALL_SAVE_FLAG);
		return SL;
	}

	/**
	 * Transfers the layer to the canvas and deletes it.
	 */
	public void TransferLayer()
	{
		if (SL > -1)
			cvs.restoreToCount(SL);
		SL = -1;
	}

	/*-----------------------------------------------------------------
	  CANVAS TRANSFORM
	 -----------------------------------------------------------------*/

	/**
	 * Rotates all subsequent drawings around the canvas center by the specified number of degrees.
	 */
	public void RotateCanvas(float Degrees)
	{
		cvs.rotate(Degrees, width >> 1, height >> 1);
	}

	/**
	 * Rotates all subsequent drawings around the pivot coordinates by the specified number of degrees.
	 */
	public void RotateCanvasAround(float Degrees, float PivotX, float PivotY)
	{
		cvs.rotate(Degrees, PivotX, PivotY);
	}

	/**
	 * Scales all subsequent drawings by ScaleX, ScaleY.
	 */
	public void ScaleCanvas(float ScaleX, float ScaleY)
	{
		cvs.scale(ScaleX, ScaleY);
	}

	/**
	 * Skews all subsequent drawings by SkewX, SkewY.
	 */
	public void SkewCanvas(float SkewX, float SkewY)
	{
		cvs.skew(SkewX, SkewY);
	}

	/**
	 * Translates all subsequent drawings by dX, dY.
	 */
	public void TranslateCanvas(float dX, float dY)
	{
		cvs.translate(dX, dY);
	}

	/*-----------------------------------------------------------------
	  CLIPPING
	 -----------------------------------------------------------------*/

	/**
	 * Clips the drawing area to the given path. The path may be an instance of the Path class or an instance of the AS_Path class.
	 * Warning: This does not work if the hardware acceleration is enabled.
	 */
	public void ClipPath(Object Path)
	{
		if (Path.getClass().getName().equals("android.graphics.Path"))
			cvs.clipPath((android.graphics.Path) Path);
		else if (Path instanceof ComplexPath)
			cvs.clipPath(((ComplexPath)Path).path);
	}

	/**
	 * Clips the drawing area to the given rectangle.
	 */
	public void ClipRect(RectWrapper Rect)
	{
		cvs.clipRect(Rect.getObject());
	}

	/**
	 * Removes the clipping set by ClipPath or ClipRect.
	 */
	public void RemoveClip()
	{
		Rect r = new Rect(0, 0, width, height);
		cvs.clipRect(r, Op.UNION);
	}

	/*-----------------------------------------------------------------
	  CAMERA
	 -----------------------------------------------------------------*/

	/**
	 * Imports a Camera from another library.
	 */
	public void ImportCamera(Camera c)
	{
		cam = c;
	}

	/**
	 * Saves the camera state. Each Save should be balanced with a call to Restore.
	 */
	public void CameraSave()
	{
		cam.save();
	}

	/**
	 * Restores the saved state, if any.
	 */
	public void CameraRestore()
	{
		cam.restore();
	}

	/**
	 * Applies a rotation transform around the X axis.
	 * @param deg The angle of rotation around the X axis, in degrees
	 */
	public void CameraRotateX(float deg)
	{
		cam.rotateX(deg);
	}

	/**
	 * Applies a rotation transform around the Y axis.
	 * @param deg The angle of rotation around the Y axis, in degrees
	 */
	public void CameraRotateY(float deg)
	{
		cam.rotateY(deg);
	}

	/**
	 * Applies a rotation transform around the Z axis.
	 * @param deg The angle of rotation around the Z axis, in degrees
	 */
	public void CameraRotateZ(float deg)
	{
		cam.rotateZ(deg);
	}

	/**
	 * Sets the location of the camera. The default location is set at 0, 0, -8.
	 * @param x The x location of the camera
	 * @param y The y location of the camera
	 * @param z The z location of the camera
	 * minSDK=12
	 */
	public void CameraSetLocation(float x, float y, float z)
	{
		cam.setLocation(x, y, z);
	}

	/**
	 * Applies a translation transform on all three axis.
	 * @param x The distance to translate by on the X axis
	 * @param y The distance to translate by on the Y axis
	 * @param z The distance to translate by on the Z axis
	 */
	public void CameraTranslate(float x, float y, float z)
	{
		cam.translate(x, y, z);
	}

	/**
	 * Computes the matrix corresponding to the current transformation and copies it to the internal matrix.
	 */
	public void CameraComputeMatrix()
	{
		cam.getMatrix(matrix);
	}
}
