package flm.b4a.accelerview;

import android.graphics.Bitmap;
import android.graphics.ComposePathEffect;
import android.graphics.CornerPathEffect;
import android.graphics.DashPathEffect;
import android.graphics.DiscretePathEffect;
import android.graphics.LightingColorFilter;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.PathDashPathEffect;
import android.graphics.PathEffect;
import android.graphics.PorterDuffColorFilter;
import android.graphics.PorterDuffXfermode;
import android.graphics.Paint.Style;
import android.graphics.PorterDuff.Mode;
import android.graphics.Rect;
import android.graphics.SumPathEffect;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.objects.drawable.CanvasWrapper.BitmapWrapper;

@BA.ShortName("AS_Object")
public class ComplexObject
{
	private boolean isInitialized = false;
	protected byte objType = -1;
	protected Bitmap bmpParam;
	protected boolean boolParam;
	protected float fParam1;
	protected float fParam2;
	protected int iWidth;
	protected int iHeight;
	protected Path pathParam;
	protected Rect rctParam;
	protected Paint paint;

	@BA.Hide
	public static final byte TYPE_ARC = 0;
	@BA.Hide
	public static final byte TYPE_BITMAP = 1;
	@BA.Hide
	public static final byte TYPE_BITMAP2 = 2;
	@BA.Hide
	public static final byte TYPE_CIRCLE = 3;
	@BA.Hide
	public static final byte TYPE_LINE = 4;
	@BA.Hide
	public static final byte TYPE_OVAL = 5;
	@BA.Hide
	public static final byte TYPE_PATH = 6;
	@BA.Hide
	public static final byte TYPE_POINT = 7;
	@BA.Hide
	public static final byte TYPE_RECT = 8;
	@BA.Hide
	public static final byte TYPE_ROUNDRECT = 9;
	@BA.Hide
	public static final byte TYPE_TEXT = 10;
	@BA.Hide
	public byte getType()
	{
		return objType;
	}

	private void Initialize(boolean Filter)
	{
		isInitialized = false;
		bmpParam = null;
		boolParam = false;
		fParam1 = 0;
		fParam2 = 0;
		iWidth = 0;
		iHeight = 0;
		pathParam = null;
		rctParam = null;
		paint = new Paint();
		paint.setAntiAlias(Filter);
		paint.setFilterBitmap(Filter);
	}

	private void CheckInitialization(String funcName) throws Exception
	{
		if (!isInitialized)
			throw(new Exception(funcName + ": object not initialized"));
	}

	/**
	 * Clones an existing AS_Object.
	 */
	public final ComplexObject Clone(ComplexObject ObjToCopy)
	{
		isInitialized = false;
		if (!ObjToCopy.isInitialized)
			return null;
		objType = ObjToCopy.objType;
		bmpParam = ObjToCopy.bmpParam;
		boolParam = ObjToCopy.boolParam;
		fParam1 = ObjToCopy.fParam1;
		fParam2 = ObjToCopy.fParam2;
		iWidth = ObjToCopy.iWidth;
		iHeight = ObjToCopy.iHeight;
		if (ObjToCopy.pathParam == null)
			pathParam = null;
		else
			pathParam = new Path(ObjToCopy.pathParam);
		if (ObjToCopy.rctParam == null)
			rctParam = null;
		else
			rctParam = new Rect(ObjToCopy.rctParam);
		if (ObjToCopy.paint == null)
			paint = null;
		else
			paint = new Paint(ObjToCopy.paint);
		isInitialized = true;
		return this;
	}

	/**
	 * Creates an arc object.
	 * @param Width,Height Size of the oval used to define the shape and size of the arc.
	 * @param StartAngle Starting angle (in degrees) where the arc begins.
	 * @param SweepAngle Sweep angle (in degrees) measured clockwise.
	 * @param Wedge Whether the center of the oval closes the arc to draw a wedge.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public final ComplexObject CreateArc(int Width, int Height, float StartAngle, float SweepAngle, boolean Wedge, boolean Filter)
	{
		Initialize(Filter);
		objType = TYPE_ARC;
		iWidth = Width;
		iHeight = Height;
		fParam1 = StartAngle;
		fParam2 = SweepAngle;
		boolParam = Wedge;
		isInitialized = true;
		return this;
	}

	/**
	 * Changes the angles of an arc object.
	 * @param StartAngle Starting angle (in degrees) where the arc begins.
	 * @param SweepAngle Sweep angle (in degrees) measured clockwise.
	 */
	public void ChangeArcAngles(float StartAngle, float SweepAngle) throws Exception
	{
		CheckInitialization("ChangeArcAngles");
		if (this.objType != TYPE_ARC)
			throw(new Exception("ChangeArcAngles: wrong object type"));
		fParam1 = StartAngle;
		fParam2 = SweepAngle;
	}

	/**
	 * Creates a bitmap object.
	 * @param Filter Filtering affects the sampling of bitmaps when they are rescaled. True = best quality.
	 */
	public final ComplexObject CreateBitmap(BitmapWrapper SrcBmp, boolean Filter)
	{
		Initialize(Filter);
		objType = TYPE_BITMAP;
		bmpParam = SrcBmp.getObject();
		isInitialized = true;
		return this;
	}

	/**
	 * Creates a bitmap object.
	 * @param SrcRect The subset of the bitmap that will be drawn. If Null then the complete bitmap will be drawn.
	 * @param Width,Height Size of the rectangle that the bitmap will be drawn into.
	 * @param Filter Filtering affects the sampling of bitmaps when they are rescaled. True = best quality.
	 */
	public final ComplexObject CreateBitmap2(BitmapWrapper SrcBmp, Rect SrcRect, int Width, int Height, boolean Filter)
	{
		Initialize(Filter);
		objType = TYPE_BITMAP2;
		bmpParam = SrcBmp.getObject();
		rctParam = SrcRect;
		iWidth = Width;
		iHeight = Height;
		isInitialized = true;
		return this;
	}

	/**
	 * Replaces the bitmap of a bitmap object.
	 */
	public void ChangeBitmap(BitmapWrapper SrcBmp) throws Exception
	{
		CheckInitialization("ChangeBitmap");
		if (this.objType != TYPE_BITMAP && this.objType != TYPE_BITMAP2)
			throw(new Exception("ChangeBitmap: wrong object type"));
		bmpParam = SrcBmp.getObject();
	}

	/**
	 * Creates a circle object.
	 * @param Radius The radius of the circle to be drawn.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public final ComplexObject CreateCircle(float Radius, boolean Filter)
	{
		Initialize(Filter);
		objType = TYPE_CIRCLE;
		fParam1 = Radius;
		isInitialized = true;
		return this;
	}

	/**
	 * Changes the radius of a circle object.
	 */
	public void ChangeCircleRadius(float Radius) throws Exception
	{
		CheckInitialization("ChangeCircleRadius");
		if (objType != TYPE_CIRCLE)
			throw(new Exception("ChangeCircleRadius: wrong object type"));
		fParam1 = Radius;
	}

	/**
	 * Creates a line object
	 * @param DeltaX,DeltaY Offsets from the starting position used to calculate the ending position.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public final ComplexObject CreateLine(float DeltaX, float DeltaY, boolean Filter)
	{
		Initialize(Filter);
		objType = TYPE_LINE;
		fParam1 = DeltaX;
		fParam2 = DeltaY;
		isInitialized = true;
		return this;
	}

	/**
	 * Creates an oval object.
	 * @param Width,Height Size of the oval to be drawn.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public final ComplexObject CreateOval(int Width, int Height, boolean Filter)
	{
		Initialize(Filter);
		objType = TYPE_OVAL;
		iWidth = Width;
		iHeight = Height;
		isInitialized = true;
		return this;
	}

	/**
	 * Creates a path object.
	 * @param Path An instance of the Path class or an instance of the AS_Path class.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public final ComplexObject CreatePath(Object Path, boolean Filter) throws Exception
	{
		Initialize(Filter);
		objType = TYPE_PATH;
		if (Path == null)
			throw(new Exception("CreatePath: Path is null"));
		if (Path.getClass().getName().equals("android.graphics.Path")) {
			pathParam = (android.graphics.Path) Path;
		}
		else if (Path instanceof ComplexPath) {
			pathParam = ((ComplexPath)Path).path;
		}
		else
			throw(new Exception("CreatePath: invalid Path"));
		isInitialized = true;
		return this;
	}

	/**
	 * Changes the path of a path object.
	 */
	public void ChangePath(Object Path) throws Exception
	{
		CheckInitialization("ChangePath");
		if (objType != TYPE_PATH)
			throw(new Exception("ChangePath: wrong object type"));
		if (Path == null)
			pathParam = null;
		else if (Path.getClass().getName().equals("android.graphics.Path"))
			pathParam = (android.graphics.Path) Path;
		else if (Path instanceof ComplexPath)
			pathParam = ((ComplexPath)Path).path;
	}

	/**
	 * Creates a point object.
	 */
	public final ComplexObject CreatePoint()
	{
		Initialize(false);
		objType = TYPE_POINT;
		isInitialized = true;
		return this;
	}

	/**
	 * Creates a rectangle object.
	 * @param Width,Height Size of the rectangle to be drawn.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public final ComplexObject CreateRect(int Width, int Height, boolean Filter)
	{
		Initialize(Filter);
		objType = TYPE_RECT;
		iWidth = Width;
		iHeight = Height;
		isInitialized = true;
		return this;
	}

	/**
	 * Creates a rectangle object with rounded corners.
	 * @param Width,Height Size of the round-rect to be drawn.
	 * @param CornerRadiusX The x-radius of the oval used to round the corners.
	 * @param CornerRadiusY The y-radius of the oval used to round the corners.
	 * @param Filter Whether the edges are smoothed out.
	 */
	public final ComplexObject CreateRoundRect(int Width, int Height, float CornerRadiusX, float CornerRadiusY, boolean Filter)
	{
		Initialize(Filter);
		objType = TYPE_ROUNDRECT;
		iWidth = Width;
		iHeight = Height;
		fParam1 = CornerRadiusX;
		fParam2 = CornerRadiusY;
		isInitialized = true;
		return this;
	}

	/**
	 * Changes the width and height of an object (if it has such properties).
	 */
	public void ChangeWidthHeight(int Width, int Height) throws Exception
	{
		CheckInitialization("ChangeWidthHeight");
		iWidth = Width;
		iHeight = Height;
	}
	/**
	 * Gets the width of an object (if it has such a property).
	 */
	public int GetWidth() throws Exception
	{
		CheckInitialization("GetWidth");
		return iWidth;
	}
	/**
	 * Gets the height of an object (if it has such a property).
	 */
	public int GetHeight() throws Exception
	{
		CheckInitialization("GetHeight");
		return iHeight;
	}

	/**
	 * Called internally by the ComplexText class. 
	 */
	protected final ComplexObject CreateText(boolean Filter)
	{
		Initialize(Filter);
		objType = TYPE_TEXT;
		isInitialized = true;
		return this;
	}

	/**
	 * Sets the alpha value of the object color, leaving its R,G,B values unchanged.
	 */
	public void SetAlpha(int Alpha) throws Exception
	{
		CheckInitialization("SetAlpha");
		paint.setAlpha(Alpha);
	}
	/**
	 * Returns the alpha value of the object color.
	 */
	public int GetAlpha() throws Exception
	{
		CheckInitialization("GetAlpha");
		return paint.getAlpha();
	}

	/**
	 * Paints the object with the given color.
	 * If this function is called after SetTexture, SetTexture is ignored. 
	 */
	public final ComplexObject SetColor(int Color) throws Exception
	{
		CheckInitialization("SetColor");
		if (paint.getShader() != null) paint.setShader(null);
		paint.setColor(Color);
		return this;
	}
	/**
	 * Gets the object color.
	 */
	public int GetColor() throws Exception
	{
		CheckInitialization("GetColor");
		return paint.getColor();
	}

	public static final Style STYLE_FILL = Style.FILL;
	public static final Style STYLE_FILL_AND_STROKE = Style.FILL_AND_STROKE;
	public static final Style STYLE_STROKE = Style.STROKE;

	/**
	 * Defines whether the object will be filled or stroked, or both.
	 * @param Style One of the STYLE constants.
	 * @param StrokeWidth The stroke width. Relevant when Style &lt;&gt; STYLE_FILL.
	 */
	public final ComplexObject SetStyle(Style style, int StrokeWidth) throws Exception
	{
		CheckInitialization("SetStyle");
		paint.setStyle(style);
		paint.setStrokeWidth(StrokeWidth);
		return this;
	}

	public static final int PATHEFFECT_CLEAR = 0;
	public static final int PATHEFFECT_DASH_ALTERNATE = 1;
	public static final int PATHEFFECT_DASH_REGULAR = 2;
	public static final int PATHEFFECT_DASH_SEGMENTED = 3;
	public static final int PATHEFFECT_DISTORTED = 4;
	public static final int PATHEFFECT_ROUNDED = 5;

	/**
	 * Sets a patheffect. This only affects drawing with a style set to STROKE or FILL_AND_STROKE.
	 * @param Effect One of the PATHEFFECT constants.
	 * @param Value Value of the effect (for dash effects, it defines the dash size and interval).
	 */
	public final ComplexObject SetPathEffect(int Effect, float Value) throws Exception
	{
		CheckInitialization("SetPathEffect");
		switch (Effect)
		{
		case PATHEFFECT_DASH_ALTERNATE:
			paint.setPathEffect(new DashPathEffect(new float[] {Value * 2.5f, Value, Value, Value}, 0));
			break;
		case PATHEFFECT_DASH_REGULAR:
			paint.setPathEffect(new DashPathEffect(new float[] {Value, Value}, 0));
			break;
		case PATHEFFECT_DASH_SEGMENTED:
			paint.setPathEffect(new DashPathEffect(new float[] {Value * 3f, Value}, 0));
			break;
		case PATHEFFECT_DISTORTED:
			paint.setPathEffect(new DiscretePathEffect(Value * 3f, Value));
			break;
		case PATHEFFECT_ROUNDED:
			paint.setPathEffect(new CornerPathEffect(Value));
			break;
		default:
			paint.setPathEffect(null);
			break;
		}
		return this;
	}

	public static final PathDashPathEffect.Style PATHEFFECT_STYLE_MORPH = PathDashPathEffect.Style.MORPH;
	public static final PathDashPathEffect.Style PATHEFFECT_STYLE_ROTATE = PathDashPathEffect.Style.ROTATE;
	public static final PathDashPathEffect.Style PATHEFFECT_STYLE_TRANSLATE = PathDashPathEffect.Style.TRANSLATE;

	/**
	 * Stamps the contour of the object with the specified shape. This only applies to drawings with a style set to STROKE or FILL_AND_STROKE.
	 * @param Shape An instance of the Path class or an instance of the AS_Path class.
	 * @param Spacing Spacing between each stamp of shape.
	 * @param Style How to transform the shape at each position as it is stamped. One of the PATHEFFECT_STYLE constants.
	 */
	public final ComplexObject SetPathEffect2(Object Shape, float Spacing, PathDashPathEffect.Style Style) throws Exception
	{
		CheckInitialization("SetPathEffect2");
		if (Shape == null)
			paint.setPathEffect(null);
		else if (Shape.getClass().getName().equals("android.graphics.Path"))
			paint.setPathEffect(new PathDashPathEffect((android.graphics.Path)Shape, Spacing, 0, Style));
		else if (Shape instanceof ComplexPath)
			paint.setPathEffect(new PathDashPathEffect(((ComplexPath)Shape).path, Spacing, 0, Style));
		return this;
	}

	private void SumPathEffect(boolean Combine, PathEffect PE)
	{
		if (Combine)
			paint.setPathEffect(new ComposePathEffect(paint.getPathEffect(), PE));
		else
			paint.setPathEffect(new SumPathEffect(paint.getPathEffect(), PE));
	}
	/**
	 * Adds a patheffect to the current patheffect. This only affects drawing with a style set to STROKE or FILL_AND_STROKE.
	 * @param Effect One of the PATHEFFECT constants.
	 * @param Value Value of the effect (for dash effects, it defines the dash size and interval).
	 * @param Combine If True, the two effects are combined. Otherwise they are just superimposed.
	 */
	public final ComplexObject AddToPathEffect(int Effect, float Value, boolean Combine) throws Exception
	{
		CheckInitialization("AddToPathEffect");
		if (paint.getPathEffect() == null)
			return SetPathEffect(Effect, Value);
		switch (Effect)
		{
		case PATHEFFECT_DASH_ALTERNATE:
			SumPathEffect(Combine, new DashPathEffect(new float[] {Value * 2.5f, Value, Value, Value}, 0));
			break;
		case PATHEFFECT_DASH_REGULAR:
			SumPathEffect(Combine, new DashPathEffect(new float[] {Value, Value}, 0));
			break;
		case PATHEFFECT_DASH_SEGMENTED:
			SumPathEffect(Combine, new DashPathEffect(new float[] {Value * 3f, Value}, 0));
			break;
		case PATHEFFECT_DISTORTED:
			SumPathEffect(Combine, new DiscretePathEffect(Value * 3f, Value));
			break;
		case PATHEFFECT_ROUNDED:
			SumPathEffect(Combine, new CornerPathEffect(Value));
			break;
		}
		return this;
	}

	/**
	 * Adds a colorfilter that multiplies the RGB channels by one color, and then adds a second color,
	 * pinning the result for each component to [0..255].
	 * If this function is called after SetPorterDuffColorFilter, SetPorterDuffColorFilter is ignored.
	 */
	public final ComplexObject SetLightingColorFilter(int Multiply, int Add) throws Exception
	{
		CheckInitialization("SetLightingColorFilter");
		paint.setColorFilter(new LightingColorFilter(Multiply, Add));
		return this;
	}

	public final static Mode MODE_CLEAR = Mode.CLEAR;
	public final static Mode MODE_DARKEN = Mode.DARKEN;
	public final static Mode MODE_LIGHTEN = Mode.LIGHTEN;
	public final static Mode MODE_MULTIPLY = Mode.MULTIPLY;
	public final static Mode MODE_SCREEN = Mode.SCREEN;
	public final static Mode MODE_SRC_ATOP = Mode.SRC_ATOP;
	public final static Mode MODE_SRC_IN = Mode.SRC_IN;
	public final static Mode MODE_SRC_OUT = Mode.SRC_OUT;
	public final static Mode MODE_SRC_OVER = Mode.SRC_OVER;
	public final static Mode MODE_DST_ATOP = Mode.DST_ATOP;
	public final static Mode MODE_DST_IN = Mode.DST_IN;
	public final static Mode MODE_DST_OUT = Mode.DST_OUT;
	public final static Mode MODE_DST_OVER = Mode.DST_OVER;
	public final static Mode MODE_XOR = Mode.XOR;

	/**
	 * Adds a colorfilter that uses the specified color and Porter-Duff mode.
	 * Warning: some modes may have a different result depending on whether the hardware acceleration is enabled or not.
	 * If this function is called after SetLightingColorFilter, SetLightingColorFilter is ignored.
	 * @param PDMode One of the MODE constants or Null.
	 */
	public final ComplexObject SetPorterDuffColorFilter(int SrcColor, Mode PDMode) throws Exception
	{
		CheckInitialization("SetPorterDuffColorFilter");
		if (PDMode == null)
			paint.setColorFilter(null);
		else
			paint.setColorFilter(new PorterDuffColorFilter(SrcColor, PDMode));
		return this;
	}

	/**
	 * Adds a transfer-mode in the drawing pipeline that uses the specified Porter-Duff mode.
	 * Warning: some modes may have a different result depending on whether the hardware acceleration is enabled or not.
	 * @param PDMode One of the MODE constants or Null.
	 */
	public final ComplexObject SetXferMode(Mode PDMode) throws Exception
	{
		CheckInitialization("SetXferMode");
		if (PDMode == null)
			paint.setXfermode(null);
		else
			paint.setXfermode(new PorterDuffXfermode(PDMode));
		return this;
	}

	/**
	 * Paints the object with the given texture.
	 * If this function is called after SetColor, SetColor is ignored. 
	 */
	public final ComplexObject SetTexture(Texture texture) throws Exception
	{
		CheckInitialization("SetTexture");
		if (texture == null || !texture.getIsInitialized())
			paint.setShader(null);
		else
			paint.setShader(texture.getShader());
		return this;
	}

	/**
	 * Gets/sets the internal Paint.
	 */
	public Paint getPaint() throws Exception
	{
		CheckInitialization("getPaint");
		return paint;
	}
	public void setPaint(Paint paint) throws Exception
	{
		CheckInitialization("setPaint");
		this.paint = paint;
	}

	public boolean getIsInitialized()
	{
		return isInitialized;
	}
}
